# -*- coding: utf-8 -*-
from rest_framework import serializers
from universal_notifications.models import Device


class DeviceSerializer(serializers.ModelSerializer):

    def create(self, data):
        data['user'] = self.context['request'].user
        return super(DeviceSerializer, self).create(data)

    class Meta:
        model = Device
        fields = ['platform', 'notification_token', 'device_id', 'app_id']



# -*- coding: utf-8 -*-
import logging

import cssutils
from django.conf import settings
from django.contrib.sites.models import Site
from django.core.mail import EmailMessage
from django.template.loader import render_to_string
from premailer import Premailer

try:
    from django.utils.importlib import import_module
except ImportError:
    from importlib import import_module

cssutils.log.setLevel(logging.CRITICAL)

if hasattr(settings, "UNSUBSCRIBE_MODEL"):
    __path, __symbol = getattr(settings, 'UNSUBSCRIBE_MODEL').rsplit('.', 1)
    UnsubscribedModel = getattr(import_module(__path), __symbol)
else:
    UnsubscribedModel = None


def send_email(template, to, subject, variables={}, fail_silently=False, cms=False, replace_variables={}):
    variables['site'] = Site.objects.get_current()
    variables['STATIC_URL'] = settings.STATIC_URL
    variables['is_secure'] = getattr(settings, 'IS_SECURE', False)
    html = render_to_string('emails/email_%s.html' % template, variables)
    protocol = 'https://' if variables['is_secure'] else 'http://'
    replace_variables['protocol'] = protocol
    domain = variables['site'].domain
    replace_variables['domain'] = domain
    for key, value in replace_variables.iteritems():
        if not value:
            value = ''
        html = html.replace('{%s}' % key.upper(), value)
    # Update path to have domains
    base = protocol + domain
    html = Premailer(html,
                     remove_classes=False,
                     exclude_pseudoclasses=False,
                     keep_style_tags=True,
                     include_star_selectors=True,
                     strip_important=False,
                     base_url=base).transform()
    email = EmailMessage(subject, html, settings.DEFAULT_FROM_EMAIL, [to])
    email.content_subtype = "html"
    email.send(fail_silently=fail_silently)



# -*- coding: utf-8 -*-
from django.conf import settings
from rest_framework.renderers import JSONRenderer
from ws4redis.publisher import RedisPublisher
from ws4redis.redis_store import RedisMessage


def publish(user, item=None, additional_data=None):
    if additional_data is None:
        additional_data = {}
    redis_publisher = RedisPublisher(facility='all', users=[user.email])
    r = JSONRenderer()
    if item is None:
        data = {}
    else:
        data = item.as_dict()
    data.update(additional_data)
    data = r.render(data)
    message = RedisMessage(data)
    if settings.TESTING:
        # Do not send in tests
        return
    redis_publisher.publish_message(message)



# -*- coding: utf-8 -*-
from django.conf.urls import url
from rest_framework.routers import DefaultRouter
from universal_notifications.api import DevicesApi

urlpatterns = [
    url(r'^devices$', DevicesApi.as_view(), name='notifications_devices_api'),
]

router = DefaultRouter()
urlpatterns += router.urls



# -*- coding: utf-8 -*-
from rest_framework.generics import CreateAPIView
from universal_notifications.serializers import DeviceSerializer


class DevicesApi(CreateAPIView):
    serializer_class = DeviceSerializer



# -*- coding: utf-8 -*-
import six
from django.conf import settings
from django.db import models
from universal_notifications.backends.push.apns import apns_send_message
from universal_notifications.backends.push.fcm import fcm_send_message
from universal_notifications.backends.push.gcm import gcm_send_message

AUTH_USER_MODEL = getattr(settings, 'AUTH_USER_MODEL', 'auth.User')


class NotificationHistory(models.Model):
    created = models.DateTimeField(auto_now_add=True)
    group = models.CharField(max_length=50)
    klass = models.CharField(max_length=255)
    receiver = models.CharField(max_length=255)
    details = models.TextField()


class Device(models.Model):
    user = models.ForeignKey(AUTH_USER_MODEL, related_name='devices')
    notification_token = models.TextField()
    device_id = models.CharField(max_length=255)
    is_active = models.BooleanField(default=True, help_text="Inactive devices will not be sent notifications")
    created = models.DateTimeField(auto_now_add=True)
    PLATFORM_IOS = 'ios'
    PLATFORM_GCM = 'gcm'
    PLATFORM_FCM = 'fcm'
    PLATFORM_CHOICES = (
        (PLATFORM_IOS, 'iOS'),
        (PLATFORM_GCM, 'Google Cloud Messagging (deprecated)'),
        (PLATFORM_FCM, 'Firebase Cloud Messaging'),
    )
    app_id = models.CharField(max_length=100)
    platform = models.CharField(max_length=10, choices=PLATFORM_CHOICES)

    def __unicode__(self):
        return "%s (%s)" % (self.user.email or "unknown user", self.device_id)

    def send_message(self, message, **data):
        """Send message to device

        Args:
            message (string): Message string
            **data (dict, optional): Extra data

        Returns:
            boolean: status of sending notification
        """
        if not self.is_active:
            return False

        if not isinstance(message, six.string_types):
            message = unicode(message)
        args = self, message, data

        if self.platform == Device.PLATFORM_GCM:
            return gcm_send_message(*args)
        elif self.platform == Device.PLATFORM_IOS:
            return apns_send_message(*args)
        elif self.platform == Device.PLATFORM_FCM:
            return fcm_send_message(*args)
        else:
            return False



# -*- coding: utf-8 -*-
from django.conf import settings

try:
    from django.utils.importlib import import_module
except ImportError:
    from importlib import import_module


__path, __symbol = getattr(settings, 'CELERY_APP_PATH').rsplit('.', 1)
app = getattr(import_module(__path), __symbol)


@app.task()
def process_chained_notification(conf, item, receivers, context, parent_result):
    """ conf - configuration of chained notification
        item, receivers, context - parameters for creating Notification subclass
        parent result - result of sending parent notification """

    notification_class = conf['class']
    transform_func = conf.get('transform_func', None)
    condition_func = conf.get('condition_func', None)

    # parameters transformation
    if transform_func:
        item, receivers, context = transform_func(item, receivers, context)

    # checking if notification should be skipped
    if condition_func:
        if not condition_func(item, receivers, context, parent_result):
            return

    # sending out notification
    notification_class(item, receivers, context).send()



# -*- coding: utf-8 -*-
__title__ = 'Universal Notification'
__version__ = '0.5.0'
__author__ = 'Pawel Krzyzaniak'
__license__ = 'Proprietary'
__copyright__ = 'Copyright 2016 Arabella'

# Version synonym
VERSION = __version__



# -*- coding: utf-8 -*-
import json
from importlib import import_module

import rest_framework_swagger as rfs
from django.apps import apps
from django.conf import settings
from django.core.exceptions import PermissionDenied
from django.shortcuts import render_to_response
from django.template import RequestContext
from django.utils import six
from django.utils.safestring import mark_safe
from django.views.generic import View
from rest_framework.settings import api_settings
from rest_framework.views import Response
from rest_framework_swagger.apidocview import APIDocView
from rest_framework_swagger.compat import OrderedDict, import_string
from rest_framework_swagger.docgenerator import DocumentationGenerator
from rest_framework_swagger.introspectors import IntrospectorHelper
from universal_notifications.notifications import (EmailNotification,
                                                   NotificationBase,
                                                   PushNotification,
                                                   SMSNotification,
                                                   WSNotification)

BASE_NOTIFICATIONS = (EmailNotification, NotificationBase, PushNotification, SMSNotification, WSNotification)


class BaseGenerator(object):
    def __init__(self, obj, *args, **kwargs):
        super(BaseGenerator, self).__init__(*args, **kwargs)
        self._obj = obj

    def get_summary(self):
        return "TODO: summary"

    def get_type(self):
        return "TODO: type"

    def get_notes(self):
        return "TODO: notes"

    def get_serializer(self):
        return None

    def skip(self):
        return False


class WSDocGenerator(BaseGenerator):
    def get_summary(self):
        return self._obj.message

    def get_type(self):
        return self._obj.serializer_class.__name__

    def get_notes(self):
        data = self._obj.serializer_class.__name__
        if self._obj.serializer_many:
            data = "[%s*]" % data
        return "<b>Message</b><br/>%s<br/><br/><b>Data</b><br/>%s" % (self._obj.message, data)

    def get_serializer(self):
        return self._obj.serializer_class

    def skip(self):
        return not self._obj.message


class SMSDocGenerator(BaseGenerator):
    def get_summary(self):
        return "SMS"

    def get_type(self):
        return None

    def get_notes(self):
        return "<b>Template:</b><br/>%s" % self._obj.message

    def skip(self):
        return not self._obj.message


class EmailDocGenerator(BaseGenerator):
    template_loader = None

    # ++ THIS IS UGLY. TODO: Find better method
    @classmethod
    def get_template_loader(cls):
        if not cls.template_loader:
            from django.template.loaders.filesystem import Loader
            from django.template.engine import Engine

            default_template_engine = Engine.get_default()
            cls.template_loader = Loader(default_template_engine)
        return cls.template_loader

    @classmethod
    def get_template(cls, template_name):
        template_loader = cls.get_template_loader()
        try:
            source, dummy = template_loader.load_template_source(template_name)
            return source
        except:
            return ''
    # --

    def get_summary(self):
        return self._obj.email_name

    def get_type(self):
        return None

    def get_notes(self):
        notes = "<b>Subject:</b><br/>%s<br/><br/><b>Preview:</b><br/>%s" % (
            self._obj.email_subject,
            self.get_template('emails/email_%s.html' % self._obj.email_name)
        )
        return mark_safe(notes)

    def skip(self):
        return not self._obj.email_name


class PatchedDocumentationGenerator(DocumentationGenerator):
    def __init__(self, *args, **kwargs):
        pass  # ignore init from oryginal documentation generator


class NotificationsDocs(object):
    _registry = {}
    _autodiscovered = False
    _serializers = set()
    _generator_mapping = {
        EmailNotification: EmailDocGenerator,
        SMSNotification: SMSDocGenerator,
        WSNotification: WSDocGenerator
    }
    _pdg = PatchedDocumentationGenerator()

    @classmethod
    def autodiscover(cls):
        if cls._autodiscovered:
            return

        cls._autodiscovered = True

        # classes
        for app_config in apps.get_app_configs():
            if app_config.name == 'universal_notifications':  # no self importing
                continue
            try:
                module = import_module('%s.%s' % (app_config.name, 'notifications'))
                for key in dir(module):
                    item = getattr(module, key)
                    if issubclass(item, NotificationBase) and item not in BASE_NOTIFICATIONS:
                        notification_type = item.get_type()
                        if notification_type not in cls._registry:
                            cls._registry[notification_type] = {}
                        item_key = "%s.%s" % (key, app_config.name)
                        item_path = "%s (%s.notifications)" % (key, app_config.name)

                        generator = cls.get_generator(item)
                        if not generator.skip():
                            serializer = generator.get_serializer()
                            if serializer:
                                cls._serializers.add(serializer)
                            cls._registry[notification_type][item_key] = {"cls": item, "path": item_path}
            except:
                pass

        # serializers
        cls._serializers.update(cls._pdg._find_field_serializers(cls._serializers))

    @classmethod
    def get_types(cls):
        return sorted(cls._registry.keys())

    @classmethod
    def get_notifications(cls, notification_type):
        return sorted(cls._registry.get(notification_type, {}).items(), key=lambda x: x[0])

    @classmethod
    def get_generator(cls, notification_cls):
        for key_cls, generator_cls in cls._generator_mapping.items():
            if issubclass(notification_cls, key_cls):
                return generator_cls(notification_cls)
        return BaseGenerator(notification_cls)

    @classmethod
    def generate_notifications_docs(cls, notification_type):
        notifications_docs = []
        for key, value in cls.get_notifications(notification_type):
            generator = cls.get_generator(value['cls'])
            notifications_docs.append({
                'operations': [{
                    'parameters': [],
                    'nickname': key,
                    'notes': generator.get_notes(),
                    'summary': generator.get_summary(),
                    'type': generator.get_type(),
                    'method': 'GET'
                }],
                'path': value['path'],
                'description': ''
            })

        return notifications_docs

    # ++ CODE TAKEN FROM DJANGO REST SWAGGER & MODIFIED

    @classmethod
    def generate_models_docs(cls):
        models_docs = {}

        for serializer in cls._serializers:
            data = cls._pdg._get_serializer_fields(serializer)

            serializer_name = IntrospectorHelper.get_serializer_name(serializer)
            r_name = serializer_name
            r_properties = OrderedDict((k, v) for k, v in data['fields'].items() if k not in data['write_only'])
            models_docs[r_name] = {
                'id': r_name,
                'required': [i for i in r_properties.keys()],
                'properties': r_properties,
            }

        return models_docs
    # -- CODE TAKEN FROM DJANGO REST SWAGGER & MODIFIED


# ++ CODE TAKEN FROM DJANGO REST SWAGGER & MODIFIED

try:
    JSONRenderer = list(filter(
        lambda item: item.format == 'json',
        api_settings.DEFAULT_RENDERER_CLASSES,
    ))[0]
except IndexError:
    from rest_framework.renderers import JSONRenderer


def get_full_base_path(request):
    try:
        base_path = rfs.SWAGGER_SETTINGS['base_path']
    except KeyError:
        return request.build_absolute_uri(request.path).rstrip('/')
    else:
        protocol = 'https' if request.is_secure() else 'http'
        return '{0}://{1}'.format(protocol, base_path.rstrip('/'))


class UniversalNotificationsUIView(View):
    def get(self, request, *args, **kwargs):
        if not self.has_permission(request):
            return self.handle_permission_denied(request)

        template_name = rfs.SWAGGER_SETTINGS.get('template_path')
        data = {
            'swagger_settings': {
                'discovery_url': "%s/api-docs/" % get_full_base_path(request),
                'api_key': rfs.SWAGGER_SETTINGS.get('api_key', ''),
                'api_version': rfs.SWAGGER_SETTINGS.get('api_version', ''),
                'token_type': rfs.SWAGGER_SETTINGS.get('token_type'),
                'enabled_methods': [],
                'doc_expansion': rfs.SWAGGER_SETTINGS.get('doc_expansion', ''),
            },
            'django_settings': {
                'CSRF_COOKIE_NAME': mark_safe(
                    json.dumps(getattr(settings, 'CSRF_COOKIE_NAME', 'csrftoken'))),
            }
        }
        return render_to_response(template_name, RequestContext(request, data))

    def has_permission(self, request):
        if rfs.SWAGGER_SETTINGS.get('is_superuser') and not request.user.is_superuser:
            return False

        if rfs.SWAGGER_SETTINGS.get('is_authenticated') and not request.user.is_authenticated():
            return False

        return True

    def handle_permission_denied(self, request):
        permission_denied_handler = rfs.SWAGGER_SETTINGS.get('permission_denied_handler')
        if isinstance(permission_denied_handler, six.string_types):
            permission_denied_handler = import_string(permission_denied_handler)

        if permission_denied_handler:
            return permission_denied_handler(request)
        else:
            raise PermissionDenied()


class UniversalNotificationsResourcesView(APIDocView):
    renderer_classes = (JSONRenderer, )

    def get(self, request, *args, **kwargs):
        apis = [{'path': '/' + path} for path in self.get_resources()]
        return Response({
            'apiVersion': rfs.SWAGGER_SETTINGS.get('api_version', ''),
            'swaggerVersion': '1.2',
            'basePath': self.get_base_path(),
            'apis': apis,
            'info': rfs.SWAGGER_SETTINGS.get('info', {
                'contact': '',
                'description': '',
                'license': '',
                'licenseUrl': '',
                'termsOfServiceUrl': '',
                'title': '',
            }),
        })

    def get_base_path(self):
        try:
            base_path = rfs.SWAGGER_SETTINGS['base_path']
        except KeyError:
            return self.request.build_absolute_uri(
                self.request.path).rstrip('/')
        else:
            protocol = 'https' if self.request.is_secure() else 'http'
            return '{0}://{1}/{2}'.format(protocol, base_path, 'api-docs')

    def get_resources(self):
        return NotificationsDocs.get_types()


class UniversalNotificationsApiView(APIDocView):
    renderer_classes = (JSONRenderer, )

    def get(self, request, path, *args, **kwargs):
        result = Response({
            'apiVersion': rfs.SWAGGER_SETTINGS.get('api_version', ''),
            'swaggerVersion': '1.2',
            'basePath': self.api_full_uri.rstrip('/'),
            'resourcePath': '/' + path,
            'apis': NotificationsDocs.generate_notifications_docs(path),
            'models': NotificationsDocs.generate_models_docs(),
        })

        return result
# -- CODE TAKEN FROM DJANGO REST SWAGGER & MODIFIED



# -*- coding: utf-8 -*-
from django.conf import settings

try:
    from django.utils.importlib import import_module
except ImportError:
    from importlib import import_module

try:
    __path, __symbol = getattr(settings, 'SEND_SMS_FUNC').rsplit('.', 1)
    send_sms = getattr(import_module(__path), __symbol)
except:
    def send_sms(to_number, text, media=None, priority=9999):
        raise NotImplementedError



# Send to single device.
from pyfcm import FCMNotification
from universal_notifications.backends.push.utils import get_app_settings


def fcm_send_message(device, message, data=None):
    app_settings = get_app_settings(device.app_id)
    api_key = app_settings.get('FCM_API_KEY')
    if not app_settings or not api_key:
        return

    push_service = FCMNotification(api_key=api_key)
    return push_service.notify_single_device(
        registration_id=device.notification_token,
        message_body=message,
        data_message=data
    )



# -*- coding: utf-8 -*-
"""
    # TODO: fill readme & TODO
    # TODO (Pawel): processors filtering (email/sms) - unsubscription, etc, mail/sms history + limiters
    # TODO (Pawel): sample transformations / conditions
    # TODO (Pawel): push (gcm, etc.) - low priority

    Sample usage:
        WSNotification(item, receivers, context).send()

    Chaining example:
        chaining = (
            {
                'class': PushNotification,  # required, must be a subclass of NotificationBase
                'delay': 0,  # required, [in seconds]
                'transform_func': None,  # optional, should take as parameters (item, receivers, context)
                                         # and return transformed item, receivers, context - see Transformations
                                         # if empty or missing - no transformation is applied
                'condition_func': None,  # optional, should take as parameters (item, receivers, context, parent_result)
                                         # and returns True if notification should be send and False if not
                                         # if function is None,
                },
        )
"""
from django.template import Context, Template
from universal_notifications.backends.emails import (UnsubscribedModel,
                                                     send_email)
from universal_notifications.backends.sms import send_sms
from universal_notifications.backends.websockets import publish
from universal_notifications.models import Device, NotificationHistory
from universal_notifications.tasks import process_chained_notification


class NotificationBase(object):
    chaining = None

    @classmethod
    def get_type(cls):
        raise NotImplementedError

    def __init__(self, item, receivers, context):
        self.item = item
        self.receivers = receivers
        self.context = context

    def prepare_receivers(self):
        raise NotImplementedError

    def prepare_message(self):
        raise NotImplementedError

    def send_inner(self, prepared_receivers, prepared_message):
        raise NotImplementedError

    def get_notification_history_details(self):
        raise NotImplementedError

    def format_receiver_for_notification_history(self, receiver):
        return receiver

    def save_notifications(self, prepared_receivers):
        for receiver in prepared_receivers:
            NotificationHistory.objects.create(
                group=self.get_type(),
                klass=self.__class__.__name__,
                receiver=self.format_receiver_for_notification_history(receiver),
                details=self.get_notification_history_details()
            )

    def send(self):
        prepared_receivers = self.prepare_receivers()
        prepared_message = self.prepare_message()
        result = self.send_inner(prepared_receivers, prepared_message)
        self.save_notifications(prepared_receivers)
        if self.chaining:
            for conf in self.chaining:
                args = [conf, self.item, self.receivers, self.context, result]
                if conf['delay'] > 0:
                    process_chained_notification.apply_async(args, countdown=conf['delay'])
                else:  # no delay, so execute function directly, not as task
                    process_chained_notification(*args)
        return result


class WSNotification(NotificationBase):
    message = None  # required
    serializer_class = None  # required, DRF serializer
    serializer_many = False

    def prepare_receivers(self):
        return set(self.receivers)

    def prepare_message(self):
        return {
            'message': self.message,
            'data': self.serializer_class(self.item, context=self.context, many=self.serializer_many).data
        }

    def format_receiver_for_notification_history(self, receiver):
        return receiver.email

    def send_inner(self, prepared_receivers, prepared_message):
        for receiver in prepared_receivers:
            publish(receiver, additional_data=prepared_message)

    def get_notification_history_details(self):
        return "message: %s, serializer: %s" % (self.message, self.serializer_class.__name__)

    @classmethod
    def get_type(cls):
        return "WebSocket"


class SMSNotification(NotificationBase):
    message = None  # required, django template string

    def prepare_receivers(self):
        return {x.phone for x in self.receivers}

    def prepare_message(self):
        return Template(self.message).render(Context({"item": self.item}))

    def send_inner(self, prepared_receivers, prepared_message):
        for receiver in prepared_receivers:
            send_sms(receiver, prepared_message)

    def get_notification_history_details(self):
        return self.prepare_message()

    @classmethod
    def get_type(cls):
        return "SMS"


class EmailNotification(NotificationBase):
    email_name = None  # required
    email_subject = None  # required

    @classmethod
    def format_receiver(cls, receiver):
        return "%s %s <%s>" % (receiver.first_name, receiver.last_name, receiver.email)

    def prepare_receivers(self):
        result = set()
        if UnsubscribedModel:
            unsubscribed_emails = UnsubscribedModel.objects.filter(
                email__in=(receiver.email for receiver in self.receivers)
            )
            unsubscribed_emails = set(unsubscribed_emails.values_list('email', flat=True))
        else:
            unsubscribed_emails = set()
        for receiver in self.receivers:
            if receiver.email not in unsubscribed_emails:
                result.add(receiver)
        return result

    def prepare_message(self):
        return {
            'item': self.item,
        }

    def format_receiver_for_notification_history(self, receiver):
        return receiver.email

    def prepare_subject(self):
        return Template(self.email_subject).render(Context({"item": self.item}))

    def send_inner(self, prepared_receivers, prepared_message):
        prepared_subject = self.prepare_subject()
        for receiver in prepared_receivers:
            prepared_message['receiver'] = receiver
            send_email(self.email_name, self.format_receiver(receiver), prepared_subject, prepared_message)

    def get_notification_history_details(self):
        return self.email_name

    @classmethod
    def get_type(cls):
        return "Email"


class PushNotification(NotificationBase):
    message = None  # required, django template string

    @classmethod
    def get_type(cls):
        return "Push"

    def prepare_receivers(self):
        return set(self.receivers)

    def prepare_body(self):  # self.item & self.context can be used here
        return {}

    def prepare_message(self):
        return {
            'message': Template(self.message).render(Context({"item": self.item})),
            'data': self.prepare_body()
        }

    def format_receiver_for_notification_history(self, receiver):
        return receiver.email

    def send_inner(self, prepared_receivers, prepared_message):  # TODO
        for receiver in prepared_receivers:
            for d in Device.objects.filter(user=receiver, is_active=True):
                d.send_message(prepared_message["message"], **prepared_message["data"])

    def get_notification_history_details(self):
        return self.prepare_message()



# -*- coding: utf-8 -*-
from django.conf.urls import include, url
from rest_framework.routers import DefaultRouter
from universal_notifications.docs import (UniversalNotificationsApiView,
                                          UniversalNotificationsResourcesView,
                                          UniversalNotificationsUIView)

urlpatterns = [
    url(r'^$', UniversalNotificationsUIView.as_view(), name="django.swagger.base.view"),
    url(r'^api-docs/$', UniversalNotificationsResourcesView.as_view(), name="django.swagger.resources.view"),
    url(r'^api-docs/(?P<path>.*)/?$', UniversalNotificationsApiView.as_view(), name='django.swagger.api.view'),

    url(r'api/', include('universal_notifications.api_urls')),
]

router = DefaultRouter()
urlpatterns += router.urls



"""
Google Cloud Messaging
Previously known as C2DM
Documentation is available on the Android Developer website:
https://developer.android.com/google/gcm/index.html
"""

import json

from push_notifications import NotificationError
from push_notifications.settings import PUSH_NOTIFICATIONS_SETTINGS as SETTINGS
from universal_notifications.backends.push.utils import get_app_settings

try:
    from urllib.request import Request, urlopen
    from urllib.parse import urlencode
except ImportError:
    # Python 2 support
    from urllib2 import Request, urlopen
    from urllib import urlencode


class GCMError(NotificationError):
    pass


def _chunks(l, n):
    """
    Yield successive chunks from list \a l with a minimum size \a n
    """
    for i in range(0, len(l), n):
        yield l[i:i + n]


def _gcm_send(app_id, data, content_type):
    app_settings = get_app_settings(app_id)
    key = app_settings.get('GCM_API_KEY')
    if not app_settings or key:
        return

    headers = {
        "Content-Type": content_type,
        "Authorization": "key=%s" % (key),
        "Content-Length": str(len(data)),
    }

    request = Request(SETTINGS["GCM_POST_URL"], data, headers)
    return urlopen(request).read()


def _gcm_send_plain(device, data, collapse_key=None, delay_while_idle=False, time_to_live=0):
    """
    Sends a GCM notification to a single registration_id.
    This will send the notification as form data.
    If sending multiple notifications, it is more efficient to use
    gcm_send_bulk_message() with a list of registration_ids
    """

    values = {"registration_id": device.notification_token}

    if collapse_key:
        values["collapse_key"] = collapse_key

    if delay_while_idle:
        values["delay_while_idle"] = int(delay_while_idle)

    if time_to_live:
        values["time_to_live"] = time_to_live

    for k, v in data.items():
        values["data.%s" % (k)] = v.encode("utf-8")

    data = urlencode(sorted(values.items())).encode("utf-8")  # sorted items for tests

    result = _gcm_send(device.app_id, data, "application/x-www-form-urlencoded;charset=UTF-8")
    if result.startswith("Error="):
        raise GCMError(result)
    return result


def _gcm_send_json(device, data, collapse_key=None, delay_while_idle=False, time_to_live=0):
    """
    Sends a GCM notification to one or more registration_ids. The registration_ids
    needs to be a list.
    This will send the notification as json data.
    """

    values = {"registration_ids": [device.notification_token]}

    if data is not None:
        values["data"] = data

    if collapse_key:
        values["collapse_key"] = collapse_key

    if delay_while_idle:
        values["delay_while_idle"] = delay_while_idle

    if time_to_live:
        values["time_to_live"] = time_to_live

    data = json.dumps(values, separators=(",", ":"), sort_keys=True).encode("utf-8")  # keys sorted for tests
    result = _gcm_send(device.app_id, data, "application/json")
    if result:
        result = json.loads(result)
        if result["failure"]:
            raise GCMError(result)
    return result


def gcm_send_message(device, message, data=None, collapse_key=None, delay_while_idle=False, time_to_live=0):
    """
    Sends a GCM notification to a single registration_id.

    This will send the notification as form data if possible, otherwise it will
    fall back to json data.
    """
    data["message"] = message
    args = data, collapse_key, delay_while_idle, time_to_live
    try:
        try:
            return _gcm_send_plain(device, *args)
        except AttributeError:
            return _gcm_send_json(device, *args)
    except GCMError:
        # TODO: check error and maybe deactivate user
        return False



# -*- coding: utf-8 -*-
from django.contrib import admin
from universal_notifications.models import NotificationHistory


class NotificationHistoryAdmin(admin.ModelAdmin):
    list_display = ('receiver', 'created', 'group', 'klass', 'details')
    readonly_fields = ('receiver', 'created', 'group', 'klass', 'details')


admin.site.register(NotificationHistory, NotificationHistoryAdmin)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import io
import os
import re

from pkg_resources import DistributionNotFound, require
from setuptools import setup


def local_open(fname):
    return open(os.path.join(os.path.dirname(__file__), fname))


def read_md(f):
    return io.open(f, 'r', encoding='utf-8').read()


def get_version(package):
    """
    Return package version as listed in `__version__` in `init.py`.
    """
    init_py = open(os.path.join(package, '__init__.py')).read()
    return re.search("__version__ = ['\"]([^'\"]+)['\"]", init_py).group(1)


def get_packages(package):
    """
    Return root package and all sub-packages.
    """
    return [dirpath
            for dirpath, dirnames, filenames in os.walk(package)
            if os.path.exists(os.path.join(dirpath, '__init__.py'))]


def get_package_data(package):
    """
    Return all files under the root package, that are not in a
    package themselves.
    """
    walk = [(dirpath.replace(package + os.sep, '', 1), filenames)
            for dirpath, dirnames, filenames in os.walk(package)
            if not os.path.exists(os.path.join(dirpath, '__init__.py'))]

    filepaths = []
    for base, filenames in walk:
        filepaths.extend([os.path.join(base, filename)
                          for filename in filenames])
    return {package: filepaths}


version = get_version('universal_notifications')


requirements = local_open('requirements/requirements-base.txt')
required_to_install = []
for dist in requirements.readlines():
    dist = dist.strip()
    try:
        require(dist)
    except DistributionNotFound:
        required_to_install.append(dist)


setup(
    name='universal_notifications',
    version=version,
    url='https://github.com/ArabellaTech/universal_notifications',
    license='MIT',
    description='High-level framework for notifications',
    long_description=read_md('README.md'),
    author='Pawel Krzyzaniak',
    author_email='pawelk@arabel.la',
    packages=get_packages('universal_notifications'),
    package_data=get_package_data('universal_notifications'),
    zip_safe=False,
    install_requires=required_to_install,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Internet :: WWW/HTTP',
    ]
)



"""
Apple Push Notification Service
Documentation is available on the iOS Developer Library:
https://developer.apple.com/library/ios/documentation/NetworkingInternet/Conceptual/RemoteNotificationsPG/Chapters/ApplePushService.html
"""

import json
import socket
import ssl
import struct
import time
from binascii import unhexlify
from contextlib import closing

from django.core.exceptions import ImproperlyConfigured
from push_notifications import NotificationError
from push_notifications.settings import PUSH_NOTIFICATIONS_SETTINGS as SETTINGS
from universal_notifications.backends.push.utils import get_app_settings


class APNSError(NotificationError):
    pass


class APNSServerError(APNSError):
    def __init__(self, status, identifier):
        super(APNSServerError, self).__init__(status, identifier)
        self.status = status
        self.identifier = identifier


class APNSDataOverflow(APNSError):
    pass


def _apns_create_socket(address_tuple, app_id):
    app_settings = get_app_settings(app_id)
    if not app_settings:
        return
    certfile = app_settings.get("APNS_CERTIFICATE")

    if not certfile:
        raise ImproperlyConfigured(
            'You need to set MOBILE_APPS[app_id]["APNS_CERTIFICATE"] to send messages through APNS.'
        )

    try:
        with open(certfile, "r") as f:
            f.read()
    except Exception as e:
        raise ImproperlyConfigured("The APNS certificate file at %r is not readable: %s" % (certfile, e))

    sock = socket.socket()
    sock = ssl.wrap_socket(sock, ssl_version=ssl.PROTOCOL_TLSv1, certfile=certfile)
    sock.connect(address_tuple)

    return sock


def _apns_create_socket_to_push(app_id):
    return _apns_create_socket((SETTINGS["APNS_HOST"], SETTINGS["APNS_PORT"]), app_id)


def _apns_create_socket_to_feedback(app_id):
    return _apns_create_socket((SETTINGS["APNS_FEEDBACK_HOST"], SETTINGS["APNS_FEEDBACK_PORT"]), app_id)


def _apns_pack_frame(token_hex, payload, identifier, expiration, priority):
    token = unhexlify(token_hex)
    # |COMMAND|FRAME-LEN|{token}|{payload}|{id:4}|{expiration:4}|{priority:1}
    frame_len = 3 * 5 + len(token) + len(payload) + 4 + 4 + 1  # 5 items, each 3 bytes prefix, then each item length
    frame_fmt = "!BIBH%ssBH%ssBHIBHIBHB" % (len(token), len(payload))
    frame = struct.pack(
        frame_fmt,
        2, frame_len,
        1, len(token), token,
        2, len(payload), payload,
        3, 4, identifier,
        4, 4, expiration,
        5, 1, priority)

    return frame


def _apns_check_errors(sock):
    timeout = SETTINGS["APNS_ERROR_TIMEOUT"]
    if timeout is None:
        return  # assume everything went fine!
    saved_timeout = sock.gettimeout()
    try:
        sock.settimeout(timeout)
        data = sock.recv(6)
        if data:
            command, status, identifier = struct.unpack("!BBI", data)
            # apple protocol says command is always 8. See http://goo.gl/ENUjXg
            assert command == 8, "Command must be 8!"
            if status != 0:
                raise APNSServerError(status, identifier)
    except socket.timeout:  # py3, see http://bugs.python.org/issue10272
        pass
    except ssl.SSLError as e:  # py2
        if "timed out" not in e.message:
            raise
    except AttributeError:
        pass
    finally:
        sock.settimeout(saved_timeout)


def _apns_send(app_id, token, alert, badge=None, sound=None, category=None, content_available=False,
               action_loc_key=None, loc_key=None, loc_args=[], extra={}, identifier=0,
               expiration=None, priority=10, socket=None):
    data = {}
    aps_data = {}

    if action_loc_key or loc_key or loc_args:
        alert = {"body": alert} if alert else {}
        if action_loc_key:
            alert["action-loc-key"] = action_loc_key
        if loc_key:
            alert["loc-key"] = loc_key
        if loc_args:
            alert["loc-args"] = loc_args

    if alert is not None:
        aps_data["alert"] = alert

    if badge is not None:
        aps_data["badge"] = badge

    if sound is not None:
        aps_data["sound"] = sound

    if category is not None:
        aps_data["category"] = category

    if content_available:
        aps_data["content-available"] = 1

    data["aps"] = aps_data
    data.update(extra)

    # convert to json, avoiding unnecessary whitespace with separators (keys sorted for tests)
    json_data = json.dumps(data, separators=(",", ":"), sort_keys=True).encode("utf-8")

    max_size = SETTINGS["APNS_MAX_NOTIFICATION_SIZE"]
    if len(json_data) > max_size:
        raise APNSDataOverflow("Notification body cannot exceed %i bytes" % (max_size))

    # if expiration isn't specified use 1 month from now
    expiration_time = expiration if expiration is not None else int(time.time()) + 2592000

    frame = _apns_pack_frame(token, json_data, identifier, expiration_time, priority)

    if socket:
        socket.write(frame)
    else:
        with closing(_apns_create_socket_to_push(app_id)) as socket:
            socket.write(frame)
            _apns_check_errors(socket)


def _apns_read_and_unpack(socket, data_format):
    length = struct.calcsize(data_format)
    data = socket.recv(length)
    if data:
        return struct.unpack_from(data_format, data, 0)
    else:
        return None


def _apns_receive_feedback(socket):
    expired_token_list = []

    # read a timestamp (4 bytes) and device token length (2 bytes)
    header_format = '!LH'
    has_data = True
    while has_data:
        try:
            # read the header tuple
            header_data = _apns_read_and_unpack(socket, header_format)
            if header_data is not None:
                timestamp, token_length = header_data
                # Unpack format for a single value of length bytes
                token_format = '%ss' % token_length
                device_token = _apns_read_and_unpack(socket, token_format)
                if device_token is not None:
                    # _apns_read_and_unpack returns a tuple, but
                    # it's just one item, so get the first.
                    expired_token_list.append((timestamp, device_token[0]))
            else:
                has_data = False
        except socket.timeout:  # py3, see http://bugs.python.org/issue10272
            pass
        except ssl.SSLError as e:  # py2
            if "timed out" not in e.message:
                raise

    return expired_token_list


def apns_send_message(device, message=None, data=None):
    """
    Sends an APNS notification to a single registration_id.
    This will send the notification as form data.

    Note that if set message should always be a string. If it is not set,
    it won't be included in the notification. You will need to pass None
    to this for silent notifications.
    """
    _apns_send(device.app_id, device.notification_token, message, **data)


def apns_fetch_inactive_ids(app_id):
    """
    Queries the APNS server for id's that are no longer active since
    the last fetch
    """
    with closing(_apns_create_socket_to_feedback(app_id)) as socket:
        inactive_ids = []
        # Maybe we should have a flag to return the timestamp?
        # It doesn't seem that useful right now, though.
        for tStamp, registration_id in _apns_receive_feedback(socket):
            inactive_ids.append(registration_id.encode('hex'))
        return inactive_ids



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('universal_notifications', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='Device',
            fields=[
                ('id', models.AutoField(auto_created=True, verbose_name='ID', serialize=False, primary_key=True)),
                ('notification_token', models.TextField()),
                ('device_id', models.CharField(max_length=255)),
                ('is_active', models.BooleanField(default=True, help_text='Inactive devices will not be sent notifications')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('app_id', models.CharField(max_length=100)),
                ('platform', models.CharField(choices=[('ios', 'iOS'), ('gcm', 'Google Cloud Messagging (deprecated)'), ('fcm', 'Firebase Cloud Messaging')], max_length=10)),
                ('user', models.ForeignKey(to=settings.AUTH_USER_MODEL, related_name='devices')),
            ],
        ),
    ]
